# dbnomics-api -- DBnomics Web API
# By: DBnomics Team <christophe.benz@cepremap.org>
#
# Copyright (C) 2017-2020 Cepremap
# https://git.nomics.world/dbnomics/dbnomics-api
#
# dbnomics-api is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# dbnomics-api is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Application definition."""


import http.client
import logging
import tempfile
from pathlib import Path

from dbnomics_data_model.storages.abstract import StorageFactory
from flasgger import Swagger
from flask import Flask
from flask_cors import CORS
from flask_matomo import Matomo
from werkzeug.middleware.profiler import ProfilerMiddleware

from . import route_handlers
from .json_errors import JsonErrors
from .version import VERSION

log = logging.getLogger(__name__)

app = Flask(__name__)
app.config.from_object("dbnomics_api.settings")

# enable Sentry if SENTRY_DSN is defined
sentry_dsn = app.config["SENTRY_DSN"]
if sentry_dsn:
    import sentry_sdk
    from sentry_sdk.integrations.flask import FlaskIntegration

    sentry_sdk.init(
        dsn=sentry_dsn,
        integrations=[FlaskIntegration()],
    )

logging.basicConfig(level=app.config["LOG_LEVEL"])

# Initialize storage factory


def create_storage_factory_from_config(config):
    storage_type = config["STORAGE_TYPE"]
    if storage_type == "filesystem":
        storage_args = {"json_data_base_dir": config["JSON_DATA_BASE_DIR"]}
    elif storage_type == "s3":
        storage_args = {
            "endpoint_url": config["S3_ENDPOINT_URL"],
            "region_name": config["S3_REGION_NAME"],
        }
    else:
        raise ValueError(f"Unsupported storage type {storage_type!r}")
    return StorageFactory.create(storage_type, storage_args)


storage_factory = create_storage_factory_from_config(app.config)

assert isinstance(app.config["DEBUG_PYTHON_HTTP_CLIENT"], bool), app.config[
    "DEBUG_PYTHON_HTTP_CLIENT"
]
if app.config["DEBUG_PYTHON_HTTP_CLIENT"]:
    http.client.HTTPConnection.debuglevel = 1

# Set Werkzeug's `strict_slashes` option for all routes instead of defining it
# for each route individually.
# From https://stackoverflow.com/a/33285603
app.url_map.strict_slashes = False

JsonErrors(app)

CORS(app)

app.config["SWAGGER"] = {
    "title": "DBnomics API",
    "description": "".join(
        (Path(__file__).parent / "api_doc.md").open().readlines()[1:]
    ),
    "basePath": app.config.get("BASE_PATH"),
    "version": VERSION,
    "uiversion": 3,
    "termsOfService": None,  # Keep `None` to prevent displaying a fake link.
}
Swagger(app)

matomo = None
matomo_auth_token = app.config["MATOMO_AUTH_TOKEN"]
matomo_base_url = app.config["MATOMO_BASE_URL"]
matomo_site_id = app.config["MATOMO_SITE_ID"]
if matomo_auth_token and matomo_base_url and matomo_site_id:
    matomo = Matomo(
        app,
        matomo_url=matomo_base_url,
        id_site=matomo_site_id,
        token_auth=matomo_auth_token,
    )

if app.config["ENABLE_PROFILER_MIDDLEWARE"]:
    tmp_dir = tempfile.TemporaryDirectory(prefix="dbnomics-api-profiler-stats-")
    log.info(
        "Profiler is enabled, stats files are written in {!r}".format(tmp_dir.name)
    )
    app.wsgi_app = ProfilerMiddleware(app.wsgi_app, profile_dir=tmp_dir.name)

app.add_url_rule("/", "index", route_handlers.index)
app.add_url_rule("/datasets/<provider_code>", "datasets", route_handlers.datasets)
app.add_url_rule(
    "/datasets/<provider_code>/<dataset_code>", "dataset", route_handlers.datasets
)
app.add_url_rule("/last-updates", "last_updates", route_handlers.last_updates)
app.add_url_rule("/providers/<provider_code>", "provider", route_handlers.provider)
app.add_url_rule("/providers", "providers", route_handlers.providers)
app.add_url_rule("/search", "search", route_handlers.search)
app.add_url_rule(
    "/series.<format>", "series_tabular_format", route_handlers.series_tabular_format
)
app.add_url_rule("/series", "series", route_handlers.series)
app.add_url_rule(
    "/series/<provider_code>/<dataset_code>",
    "series_of_dataset",
    route_handlers.series_of_dataset,
)
app.add_url_rule(
    "/series/<provider_code>/<dataset_code>/<series_code>",
    "series_of_dataset",
    route_handlers.series_of_dataset,
)
