# dbnomics-api -- DBnomics Web API
# By: DBnomics Team <christophe.benz@cepremap.org>
#
# Copyright (C) 2017, 2018 Cepremap
# https://git.nomics.world/dbnomics/dbnomics-api
#
# dbnomics-api is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# dbnomics-api is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Functions manipulating providers, datasets and series in JSON."""

import dbnomics_data_model.observations
from dbnomics_data_model.datasets import (
    get_dimensions_codes_order,
    get_dimensions_values_labels,
)
from dbnomics_data_model.exceptions import StorageError
from dbnomics_data_model.observations import (
    normalize_observations,
    period_to_start_day,
    start_day_to_period,
)
from dbnomics_data_model.series import generate_series_name
from toolz import groupby, merge

from . import solr
from .utils import omit, without_none_values

# EXCEPTIONS


class SeriesIdsResolved(StopIteration):
    """Some releases have been resolved.

    Interrupt a function, asking for a redirection because of a release code such as "latest".
    """

    def __init__(self, series_ids):
        self.series_ids = series_ids
        super().__init__()

    def __repr__(self):
        return f"{self.__class__.__name__}(series_ids={self.series_ids})"


class SeriesPeriodError(Exception):
    def __init__(self, series_id):
        self.series_id = series_id
        super().__init__(f"Could not parse periods of series {series_id!r}")


# PROVIDERS


def build_provider(provider_json, provider_solr=None):
    """Return a `dict` describing a provider, built from JSON metadata and completed by Solr document."""
    if provider_solr is None:
        return provider_json

    return merge(omit(solr.INTERNAL_FIELDS, provider_solr), provider_json)


# DATASETS


def build_dataset(dataset_json, dataset_solr=None):
    """Return a `dict` describing a dataset, built from JSON metadata and completed by Solr document."""
    if "series" in dataset_json:
        # Remove `series` to keep only meta-data of dataset.
        dataset_json = omit(["series"], dataset_json)

    if dataset_solr is None:
        return dataset_json

    return merge(omit(solr.INTERNAL_FIELDS, dataset_solr), dataset_json)


# SERIES


def add_observations(series, observations):
    """Add attributes to `series` `dict` describing observations."""
    new_attributes = {}

    frequency, normalized_observations = normalize_observations(observations)

    new_attributes["@frequency"] = frequency

    if len(normalized_observations) >= 2:  # at least a header and one row
        transposed_observations = list(zip(*normalized_observations))
        periods = list(transposed_observations[0][1:])
        new_attributes["period"] = periods
        period_start_days = list(map(period_to_start_day, periods))
        if None in period_start_days:
            series_id = "/".join(
                [series["provider_code"], series["dataset_code"], series["series_code"]]
            )
            raise SeriesPeriodError(series_id)
        new_attributes["period_start_day"] = period_start_days
        new_attributes["value"] = list(transposed_observations[1][1:])

        # Observations attributes
        header = normalized_observations[0]
        rows = normalized_observations[1:]
        # Implement as list to keep order or columns.
        observations_attributes = [
            (column_name, list_or_value([t[index] for t in rows]))
            for index, column_name in enumerate(
                header[2:], start=2
            )  # Skip period and value, already processed.
        ]
        if observations_attributes:
            series["observations_attributes"] = observations_attributes

    return merge(series, new_attributes)


def build_series(provider_code, dataset_json, series_solr):
    """Return a `dict` describing a series, built from Solr document for performance reasons."""
    dataset_code = dataset_json["code"]
    series_code = series_solr["code"]

    series = without_none_values(
        {
            "dataset_code": dataset_code,
            "dataset_name": dataset_json.get("name"),
            "indexed_at": series_solr.get("indexed_at"),
            "provider_code": provider_code,
            "series_code": series_code,
        }
    )

    dimensions = dict(solr.iter_series_solr_dimensions(series_solr))
    if dimensions:
        series["dimensions"] = dimensions

    # Generate a name if needed
    series_name = series_solr.get("name")
    if series_name is None and dimensions:
        dimensions_codes_order = get_dimensions_codes_order(dataset_json)
        dimensions_values_labels = get_dimensions_values_labels(dataset_json)
        series_name = generate_series_name(
            dimensions, dimensions_codes_order, dimensions_values_labels
        )
    if series_name is not None:
        series["series_name"] = series_name

    return series


def build_series_list(series_ids, series_solr_by_id, observations):
    from . import app  # Avoid cyclic imports

    provider_by_code = {}
    dataset_by_id = {}
    errors = []

    if not series_ids:
        return (provider_by_code, dataset_by_id, [], errors)

    series_by_id = {}
    for (
        storage,
        series_codes_by_dataset_code,
        storage_error,
    ) in app.storage_factory.iter_series_by_storage(series_ids):
        if storage_error is not None:
            errors.append(
                error_from_storage_error(storage_error, "Could not load storage")
            )
            continue

        # Provider
        provider_code = storage.provider_code
        try:
            provider_json = storage.load_provider_json()
        except StorageError as exc:
            errors.append(error_from_storage_error(exc, "Could not load provider"))
            continue
        provider_solr = solr.get_provider(provider_code)
        provider = build_provider(provider_json, provider_solr)
        provider_by_code[provider_code] = provider

        # Handle "latest" release codes in series_ids.
        releases = storage.load_releases()
        if releases is not None:
            are_some_releases_resolved = False
            resolved_series_ids = []
            for dataset_code, series_codes in series_codes_by_dataset_code.items():
                resolved_dataset_code = releases.resolve_release_code(dataset_code)
                for series_code in series_codes:
                    resolved_series_ids.append(
                        (provider_code, resolved_dataset_code, series_code)
                    )
                if resolved_dataset_code != dataset_code:
                    are_some_releases_resolved = True
            if are_some_releases_resolved:
                raise SeriesIdsResolved(resolved_series_ids)

        for dataset_code, series_codes in series_codes_by_dataset_code.items():
            # Dataset
            dataset_dir = storage.load_dataset_dir(dataset_code)
            try:
                dataset_json = dataset_dir.load_dataset_json()
            except StorageError as exc:
                errors.append(error_from_storage_error(exc, "Could not load dataset"))
                continue
            dataset_solr = solr.get_dataset(provider_code, dataset_code)
            dataset = build_dataset(dataset_json, dataset_solr)
            dataset_by_id["{}/{}".format(provider_code, dataset_code)] = dataset

            series_list, series_list_errors = build_series_list_for_dataset(
                dataset_dir, dataset_json, series_codes, series_solr_by_id, observations
            )
            if series_list_errors:
                errors.extend(series_list_errors)
            for series in series_list:
                # Series
                series_by_id[get_series_id(series)] = series

    # Reconstitute list keeping order given in `series_ids`.
    series_list = []
    for series_id in series_ids:
        series = series_by_id.get(series_id)
        if series is not None:
            series_list.append(series)

    return (provider_by_code, dataset_by_id, series_list, errors)


def build_series_list_for_dataset(
    dataset_dir, dataset_json, series_codes, series_solr_by_id, observations
):
    errors = []

    if not series_codes:
        return ([], errors)

    provider_code = dataset_dir.storage.provider_code
    dataset_code = dataset_json["code"]

    series_by_code = {}

    for series_code in series_codes:
        series_id = dataset_dir.series_id(series_code)
        series_solr = series_solr_by_id.get(series_id)
        if series_solr is None:
            errors.append(
                {
                    "message": "Could not load series",
                    "provider_code": provider_code,
                    "dataset_code": dataset_code,
                    "series_code": series_code,
                }
            )
            continue
        series = build_series(provider_code, dataset_json, series_solr)
        series_by_code[series_code] = series

    if observations:
        found_series_code = series_by_code.keys()
        for exc, series_code, observations1 in dataset_dir.iter_observations(
            found_series_code
        ):
            if exc:
                # TODO Send exception to error logging application.
                errors.append(
                    {
                        "message": "Could not load series observations",
                        "provider_code": provider_code,
                        "dataset_code": dataset_code,
                        "series_code": series_code,
                    }
                )
            else:
                series = series_by_code[series_code]
                series = add_observations(series, observations1)
                series_by_code[series_code] = series

    # Reconstitute list keeping order given in `series_codes`.
    series_list = []
    for series_code in series_codes:
        series = series_by_code.get(series_code)
        if series is not None:
            series_list.append(series)

    return (series_list, errors)


def align_series_list(series_list):
    """Align many series along the same periods.

    See also: https://db.nomics.world/docs/web-api/#time-series-alignment
    """
    if not series_list:
        return []
    if len(series_list) == 1:
        return series_list

    aligned_series_by_id = {}
    series_list_by_frequency = groupby(
        lambda series: series.get("@frequency"), series_list
    )
    # Iterate series by frequencies and align periods
    for frequency, series_sublist in series_list_by_frequency.items():
        if frequency is None:
            # Can't align series if frequency is unknown.
            continue

        # Align series of this frequency by using datamodel align_series_list()
        aligned_series_list = dbnomics_data_model.observations.align_series_list(
            [
                (series["period_start_day"], series["value"])
                for series in series_sublist
            ],
        )
        # Iterate result and compute period as string from aligned period dates.
        for series, (completed_periods_dates, values) in zip(
            series_sublist, aligned_series_list
        ):
            aligned_series_by_id[get_series_id(series)] = merge(
                series,
                {
                    "period": [
                        start_day_to_period(completed_period_date, frequency=frequency)
                        for completed_period_date in completed_periods_dates
                    ],
                    "period_start_day": completed_periods_dates,
                    "value": values,
                },
            )

    return [
        aligned_series_by_id.get(get_series_id(series), series)
        for series in series_list
    ]


def get_series_id(series):
    return (series["provider_code"], series["dataset_code"], series["series_code"])


def list_or_value(values):
    """Return a single value if all the values are identical, or values.

    >>> list_or_value([])
    []
    >>> list_or_value([1])
    1
    >>> list_or_value([1, 1, 1])
    1
    >>> list_or_value([1, 2, 3])
    [1, 2, 3]
    """
    if not values:
        return values
    return values[0] if len(set(values)) == 1 else values


def series_to_json(series):
    new_attributes = {}
    frequency = series.get("@frequency")
    if frequency is not None:
        new_attributes["@frequency"] = frequency.value
    period_start_day = series.get("period_start_day")
    if period_start_day is not None:
        new_attributes["period_start_day"] = [
            date.isoformat() for date in period_start_day
        ]
    return merge(series, new_attributes)


# ERRORS


def error_from_storage_error(exc, message):
    return {
        "message": message,
        "cause": exc.base_message,
        "provider_code": exc.provider_code,
        "dataset_code": exc.dataset_code,
        "series_code": exc.series_code,
    }
