# dbnomics-api -- DBnomics Web API
# By: DBnomics Team <christophe.benz@cepremap.org>
#
# Copyright (C) 2017, 2018 Cepremap
# https://git.nomics.world/dbnomics/dbnomics-api
#
# dbnomics-api is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# dbnomics-api is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Functions related to series codes masks."""


import re

from toolz import mapcat


def series_code_mask_to_dimensions(code_mask, dimensions_codes):
    """Transform a series code mask to a dimensions dict.

    >>> series_code_mask_to_dimensions('', [])
    {}
    >>> series_code_mask_to_dimensions('', ['freq', 'indicator', 'geo'])
    {}
    >>> series_code_mask_to_dimensions('M.MIO_M3.AL', ['freq', 'indicator', 'geo'])
    {'freq': ['M'], 'indicator': ['MIO_M3'], 'geo': ['AL']}
    >>> series_code_mask_to_dimensions('.MIO_M3.AL', ['freq', 'indicator', 'geo'])
    {'indicator': ['MIO_M3'], 'geo': ['AL']}
    >>> series_code_mask_to_dimensions('M..AT+BE', ['freq', 'indicator', 'geo'])
    {'freq': ['M'], 'geo': ['AT', 'BE']}
    >>> series_code_mask_to_dimensions('M.MIO_M3.', ['freq', 'indicator', 'geo'])
    {'freq': ['M'], 'indicator': ['MIO_M3']}
    >>> series_code_mask_to_dimensions('M.MIO_M3', ['freq', 'indicator', 'geo'])
    {'freq': ['M'], 'indicator': ['MIO_M3']}

    # dimensions containing "+" protected by double quotes
    >>> series_code_mask_to_dimensions('"M+A"..AT+BE', ['freq', 'indicator', 'geo'])
    {'freq': ['M+A'], 'geo': ['AT', 'BE']}
    >>> series_code_mask_to_dimensions('X+"M+A"..BE', ['freq', 'indicator', 'geo'])
    {'freq': ['X', 'M+A'], 'geo': ['BE']}
    """

    def cut_dimension_expression(expression):
        """Cut one dimension expression onto dimensions values.
        >>> cut_dimension_expression('A')
        ['A']
        >>> cut_dimension_expression('A+B')
        ['A', 'B']
        >>> cut_dimension_expression('A B')
        ['A', 'B']
        >>> cut_dimension_expression('A+"C+D"')
        ['A','C+D']
        >>> cut_dimension_expression('A "C+D"+"X+Z" F')
        ['A', 'C+D', 'X+Z', 'F']
        """

        def is_double_quoted(s):
            """Return True if s is a double-quoted string (ie: '"A"')"""
            return s.startswith('"') and s.endswith('"')

        if not '"' in expression and not "+" in expression:
            return [expression]
        # Isolate double quoted parts
        chunks = re.split(
            r'\+?(".+?")\+?', expression
        )  # '"A+B"+C+D' => ['', '"A+B"', 'C+D']
        # Remove empty chunks
        chunks = filter(None, chunks)  # ['', '"A+B"', 'C+D'] => ['"A+B"', 'C+D']
        # Split each non-quoted chunks around '+'
        chunks = mapcat(
            lambda e: e.split("+") if not is_double_quoted(e) else [e], chunks
        )  # ['"A+B"', 'C+D'] => ['"A+B"', 'C', 'D']
        # remove double-quotes around double-quoted items
        # ['"A+B"', 'C', 'D'] => ['A+B', 'C', 'D']
        chunks = map(lambda e: e if not is_double_quoted(e) else e[1:-1], chunks)
        return list(chunks)

    # Split around dots
    fragments = code_mask.split(".")
    fragments = list(
        map(lambda e: cut_dimension_expression(e) if e else None, fragments)
    )
    return {
        dimension_code: dimensions_values_codes
        for dimension_code, dimensions_values_codes in zip(dimensions_codes, fragments)
        if dimensions_values_codes
    }
