# dbnomics-api -- DBnomics Web API
# By: DBnomics Team <christophe.benz@cepremap.org>
#
# Copyright (C) 2017, 2018 Cepremap
# https://git.nomics.world/dbnomics/dbnomics-api
#
# dbnomics-api is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# dbnomics-api is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Application route handlers."""


import logging
import urllib
from collections import OrderedDict
from typing import Optional

import solrq
import valideer as V
from dbnomics_data_model.datasets import (
    get_dimensions_codes_order,
    get_dimensions_values_labels,
)
from dbnomics_data_model.exceptions import StorageError
from dbnomics_data_model.releases import NoReleaseError
from flasgger import swag_from
from flask import redirect, request, url_for
from toolz import merge, pluck

from . import masks, solr, storages
from .json_errors import abort_json, make_json_response
from .json_responses import (
    SeriesIdsResolved,
    align_series_list,
    build_dataset,
    build_series_list,
    build_series_list_for_dataset,
    series_to_json,
)
from .tabular_responses import (
    TABULAR_FORMATS,
    respond_download_spreadsheet_for_different_datasets,
    respond_download_spreadsheet_for_same_dataset,
)
from .utils import normalize_string, omit, pick
from .validation import (
    check_bool,
    check_dimensions,
    check_is_positive,
    check_offset,
    check_series_ids,
    check_series_observations_args,
    convert_args_or_abort,
    guess_bool,
    series_id_str_to_tuple,
    series_observations_args,
)

log = logging.getLogger(__name__)


# HELPERS


def get_resolved_dataset_code(dataset_code: str, storage) -> Optional[str]:
    """Resolve release code in dataset code.

    Do nothing if releases are not defined for this storage.
    """
    releases = storage.load_releases()

    if releases is None:
        return None

    resolved_dataset_code = releases.resolve_release_code(dataset_code)

    if resolved_dataset_code != dataset_code:
        return resolved_dataset_code

    return None


def series_ids_to_str(series_ids):
    return ",".join("/".join(series_id) for series_id in series_ids)


# JSON PARTS OF RESPONSES


def build_dimensions_facets(
    response_facets, dimensions_codes_order, dimensions_values_labels
):
    ordered_dimensions_values = build_ordered_dimensions_values(
        dimensions_values_labels
    )
    ordered_dimensions_values_codes = {
        dimension_code: [code for code, _ in dimensions_values_tuples]
        for dimension_code, dimensions_values_tuples in ordered_dimensions_values.items()  # noqa
    }
    dimensions_facets = {}
    for dimension_code in dimensions_codes_order:
        bucketdict = {
            d["val"]: d["count"]
            for d in response_facets["dimension.{}".format(dimension_code)]["buckets"]
        }
        dimensions_facets[dimension_code] = [
            {
                "code": code,
                "count": bucketdict[code],
            }
            for code in ordered_dimensions_values_codes.get(
                dimension_code, sorted(bucketdict.keys())
            )
            if code in bucketdict
        ]
    return dimensions_facets


def build_ordered_dimensions_values(dimensions_values_labels):
    """Return a `dict` like `{'FREQ': [('M', 'Monthly')]}` where each list of
    `(dimension_value_code, dimension_value_label)` is sorted by
    dimension_value_label`.

    `dimensions_values_labels` values can be `dict` or `OrderedDict`.
    """
    ordered_dimensions_values = {}
    for dimension_code, values in dimensions_values_labels.items():
        if isinstance(values, OrderedDict):
            values_tuples = list(values.items())
        else:
            assert isinstance(values, dict), values
            # If dict, sort on normalized (lowercase, unaccentized) labels
            values_tuples = sorted(values.items(), key=lambda t: normalize_string(t[1]))
        ordered_dimensions_values[dimension_code] = values_tuples
    return ordered_dimensions_values


def build_paginated_data(docs, limit, num_found, offset):
    return {
        "docs": docs,
        "limit": limit,
        "num_found": num_found,
        "offset": offset,
    }


# ROUTE HANDLERS


def index():
    from .app import app

    apidocs_href = "{}/apidocs".format(app.config.get("BASE_PATH") or "")
    apidocs_url = urllib.parse.urljoin(request.url, apidocs_href)
    return make_json_response(
        {
            "apidocs_href": apidocs_href,
            "message": "This is the home page of DBnomics Web API. "
            "Its documentation is here: {}".format(apidocs_url),
        },
        args=None,
    )


@swag_from("openapi_specs/datasets.yml")
def datasets(provider_code, dataset_code=None):
    args = convert_args_or_abort(
        request.args,
        {
            "limit": V.Nullable(
                V.Range(V.AdaptTo(int), min_value=0, max_value=500), default=50
            ),
            "offset": check_offset,
        },
    )

    storage = storages.init_storage_or_abort(args, provider_code)

    if dataset_code is not None:
        try:
            resolved_dataset_code = get_resolved_dataset_code(dataset_code, storage)
        except {NoReleaseError, StorageError} as exc:
            abort_json(400, args, str(exc))
        if resolved_dataset_code is not None:
            return redirect(
                url_for(
                    ".dataset",
                    provider_code=provider_code,
                    dataset_code=resolved_dataset_code,
                    **request.args,  # Keep original query string
                ),
                code=302,
            )

    datasets_response = solr.find_datasets_by_provider_code(
        provider_code,
        dataset_code,
        limit=args["limit"],
        offset=args["offset"],
        sort="code ASC",
    )

    datasets = []
    for dataset_solr in datasets_response["response"]["docs"]:
        dataset_code = dataset_solr["code"]
        dataset_dir = storage.load_dataset_dir(dataset_code)
        dataset_json = storages.load_dataset_json_or_abort(
            404, args, storage, dataset_code, dataset_dir
        )
        dataset = build_dataset(dataset_json, dataset_solr)
        datasets.append(dataset)

    return make_json_response(
        {
            "datasets": build_paginated_data(
                datasets,
                args["limit"],
                datasets_response["response"]["numFound"],
                args["offset"],
            ),
        },
        args,
    )


@swag_from("openapi_specs/last_updates.yml")
def last_updates():
    args = convert_args_or_abort(
        request.args,
        {
            "datasets.limit": V.Nullable(
                V.Range(V.AdaptTo(int), min_value=0, max_value=100), default=100
            ),
            "datasets.offset": check_offset,
            "providers.limit": V.Nullable(
                V.Range(V.AdaptTo(int), min_value=0, max_value=100), default=100
            ),
            "providers.offset": check_offset,
            "sort": V.Nullable(
                V.Enum(["converted_at", "created_at"]), default="converted_at"
            ),
        },
    )

    sort = "{} DESC".format(args["sort"])

    datasets_response = solr.find_last_updated(
        type="dataset",
        limit=args["datasets.limit"],
        offset=args["datasets.offset"],
        sort=sort,
    )
    datasets = [
        omit(solr.INTERNAL_FIELDS, dataset_solr)
        for dataset_solr in datasets_response["response"]["docs"]
    ]

    providers_response = solr.find_last_updated(
        type="provider",
        limit=args["providers.limit"],
        offset=args["providers.offset"],
        sort=sort,
    )
    providers = [
        omit(solr.INTERNAL_FIELDS, provider_solr)
        for provider_solr in providers_response["response"]["docs"]
    ]

    return make_json_response(
        {
            "datasets": build_paginated_data(
                datasets,
                args["datasets.limit"],
                datasets_response["response"]["numFound"],
                args["datasets.offset"],
            ),
            "providers": build_paginated_data(
                providers,
                args["providers.limit"],
                providers_response["response"]["numFound"],
                args["providers.offset"],
            ),
        },
        args,
    )


@swag_from("openapi_specs/provider.yml")
def provider(provider_code):
    args = {"provider_code": provider_code}

    provider_solr = solr.get_provider(provider_code)
    if provider_solr is None:
        abort_json(404, args, "Provider {!r} not found".format(provider_code))
    provider_solr = omit(solr.INTERNAL_FIELDS, provider_solr)

    storage = storages.init_storage_or_abort(args, provider_code)
    category_tree = storage.load_category_tree_json()
    if category_tree is None:
        datasets_response = solr.find_datasets_by_provider_code(
            provider_code, dataset_code=None, limit=1000, offset=0, sort="name ASC"
        )
        datasets_solr = datasets_response["response"]["docs"]
        category_tree = [
            pick(["code", "name"], dataset_solr) for dataset_solr in datasets_solr
        ]

    return make_json_response(
        {
            "category_tree": category_tree,
            "provider": provider_solr,
        },
        args,
    )


@swag_from("openapi_specs/providers.yml")
def providers():
    args = convert_args_or_abort(
        request.args,
        {
            "limit": V.Nullable(check_is_positive, default=1000),
            "offset": check_offset,
        },
    )

    providers_response = solr.find_providers(limit=args["limit"], offset=args["offset"])
    providers = [
        omit(solr.INTERNAL_FIELDS, provider_solr)
        for provider_solr in providers_response["response"]["docs"]
    ]

    datasets_response = solr.find_datasets(limit=0, offset=0)
    series_response = solr.find_series(limit=0, offset=0)

    return make_json_response(
        {
            "nb_datasets": datasets_response["response"]["numFound"],
            "nb_series": series_response["response"]["numFound"],
            "providers": build_paginated_data(
                providers,
                args["limit"],
                providers_response["response"]["numFound"],
                args["offset"],
            ),
        },
        args,
    )


def escape_user_search_query(q):
    """Escape user search query to disable Solr syntax.
    If it contains only spaces, return `None`, interpreted as "all results".
    """
    return None if q.strip() == "" else solrq.Value(q)


@swag_from("openapi_specs/search.yml")
def search():
    from .app import app

    args = convert_args_or_abort(
        request.args,
        {
            "limit": V.Nullable(
                V.Range(
                    V.AdaptTo(int),
                    min_value=0,
                    max_value=app.config["SEARCH_RESULTS_LIMIT"],
                ),
                default=10,
            ),
            "offset": check_offset,
            "q": V.Nullable(str, default=""),
        },
    )

    # Escape unsafe user query
    user_query = escape_user_search_query(args["q"])

    if user_query is None:
        datasets_response = solr.find_datasets(
            limit=args["limit"], offset=args["offset"], sort="id ASC"
        )
        datasets_solr = datasets_response["response"]["docs"]
        num_found = datasets_response["response"]["numFound"]
    else:
        datasets_solr, facets_count = solr.find_datasets_from_series(
            user_query, args["limit"], args["offset"]
        )
        num_found = facets_count
    datasets_solr = [
        omit(solr.INTERNAL_FIELDS, dataset_solr) for dataset_solr in datasets_solr
    ]

    return make_json_response(
        {
            "results": build_paginated_data(
                datasets_solr,
                limit=args["limit"],
                num_found=num_found,
                offset=args["offset"],
            ),
        },
        args,
    )


def series_tabular_format(format):
    return redirect(url_for(".series", **{**request.args, "format": format}))


@swag_from("openapi_specs/series.yml")
def series():
    from .app import app

    args = convert_args_or_abort(
        request.args,
        merge(
            {
                "facets": check_bool,
                "limit": V.Nullable(
                    V.Range(
                        V.AdaptTo(int),
                        min_value=0,
                        max_value=app.config["SERIES_LIMIT"],
                    ),
                    default=app.config["SERIES_LIMIT"],
                ),
                "offset": check_offset,
                "series_ids": check_series_ids(app.config["SERIES_LIMIT"]),
                "metadata": V.Nullable(V.AdaptBy(guess_bool), default=True),
            },
            series_observations_args,
        ),
    )
    args = check_series_observations_args(args)
    if len(args["series_ids"]) > args["limit"]:
        abort_json(
            400,
            args,
            "{} 'series_ids' given but 'limit' is {}".format(
                len(args["series_ids"]), args["limit"]
            ),
        )

    series_response = solr.find_series(
        limit=args["limit"],
        offset=args["offset"],
        compute_facets=False,
        series_ids=args["series_ids"],
    )

    series_solr_by_id = {
        series_id_str_to_tuple(series_solr["id"]): series_solr
        for series_solr in series_response["response"]["docs"]
    }

    if args["format"] in TABULAR_FORMATS:
        return respond_download_spreadsheet_for_different_datasets(
            args, series_solr_by_id
        )

    try:
        (
            provider_by_code,
            dataset_by_id,
            series_list,
            series_list_errors,
        ) = build_series_list(
            args["series_ids"], series_solr_by_id, args["observations"]
        )
    except SeriesIdsResolved as exc:
        return redirect(
            url_for(
                ".series",
                **omit({"series_ids"}, request.args),  # Keep original query string
                series_ids=series_ids_to_str(
                    exc.series_ids
                ),  # Overwrite series_ids with resolved ones
            ),
            code=302,
        )

    if args["align_periods"]:
        assert args["observations"]  # checked in query string parameters validation
        series_list = align_series_list(series_list)

    response = {
        "errors": series_list_errors or None,
        "series": build_paginated_data(
            [series_to_json(series) for series in series_list],
            args["limit"],
            series_response["response"]["numFound"],
            offset=args["offset"],
        ),
    }

    if args["metadata"]:
        response["providers"] = provider_by_code
        response["datasets"] = dataset_by_id

    return make_json_response(response, args)


@swag_from("openapi_specs/series_of_dataset.yml")
def series_of_dataset(provider_code, dataset_code, series_code=None):
    from .app import app

    args = convert_args_or_abort(
        request.args,
        merge(
            {
                "dimensions": check_dimensions,
                "facets": check_bool,
                "limit": V.Nullable(
                    V.Range(
                        V.AdaptTo(int),
                        min_value=0,
                        max_value=app.config["SERIES_LIMIT"],
                    ),
                    default=app.config["SERIES_LIMIT"],
                ),
                "offset": check_offset,
                "q": V.Nullable(str, default=""),
                "metadata": V.Nullable(V.AdaptBy(guess_bool), default=True),
            },
            series_observations_args,
        ),
    )
    args = check_series_observations_args(args)
    if "dimensions" in request.args and series_code:
        abort_json(
            400,
            args,
            "'dimensions' and 'series_code' parameters can't be used together",
        )
    args = merge(
        args,
        {
            "dataset_code": dataset_code,
            "provider_code": provider_code,
            "series_code": series_code,
        },
    )

    # apply TABULAR_FORMATS presets to handle .csv and .xlsx routes
    for tabular_format in TABULAR_FORMATS:
        if dataset_code.endswith(tabular_format):
            dataset_code = dataset_code[: -len(f".{tabular_format}")]
            args["format"] = tabular_format
            args["observations"] = True
            args["align_periods"] = True
        elif series_code and series_code.endswith(tabular_format):
            series_code = series_code[: -len(f".{tabular_format}")]
            args["format"] = tabular_format
            args["observations"] = True

    storage = storages.init_storage_or_abort(args, provider_code)

    try:
        resolved_dataset_code = get_resolved_dataset_code(dataset_code, storage)
    except {NoReleaseError, StorageError} as exc:
        abort_json(400, args, str(exc))
    if resolved_dataset_code is not None:
        return redirect(
            url_for(
                ".series_of_dataset",
                provider_code=provider_code,
                dataset_code=resolved_dataset_code,
                series_code=series_code,
                **request.args,  # Keep original query string
            ),
            code=302,
        )

    provider_solr = solr.get_provider(provider_code)
    if provider_solr is None:
        abort_json(404, args, "Provider {!r} not found".format(provider_code))
    provider_solr = omit(solr.INTERNAL_FIELDS, provider_solr)

    dataset_solr = solr.get_dataset(provider_code, dataset_code)
    if dataset_solr is None:
        abort_json(
            404,
            args,
            "Dataset {!r} not found".format("/".join([provider_code, dataset_code])),
        )

    dataset_dir = storage.load_dataset_dir(dataset_code)
    dataset_json = storages.load_dataset_json_or_abort(
        404, args, storage, dataset_code, dataset_dir
    )
    dimensions_codes_order = get_dimensions_codes_order(dataset_json)

    # Escape unsafe user query
    user_query = escape_user_search_query(args["q"])

    dimensions = None
    if series_code is None:
        dimensions = args["dimensions"]
    else:
        # First, try considering `series_code` a code.
        series_id = [provider_code, dataset_code, series_code]
        series_response = solr.find_series(
            limit=args["limit"],
            offset=args["offset"],
            compute_facets=args["facets"],
            query=user_query,
            provider_code=provider_code,
            dataset_code=dataset_code,
            series_ids=[series_id],
            dimensions_codes_order=dimensions_codes_order,
        )
        if series_response["response"]["numFound"] == 1:
            series_solr = series_response["response"]["docs"][0]
            series_solr_by_id = {tuple(series_id): series_solr}
            series_codes = [series_code]
        else:
            # Try considering `series_code` a mask.
            assert series_response["response"]["numFound"] == 0, series_response
            dimensions = masks.series_code_mask_to_dimensions(
                series_code, dimensions_codes_order
            )
            log.debug(
                "Using dimensions %r from series code mask %r", dimensions, series_code
            )

    if dimensions is not None:
        series_response = solr.find_series(
            limit=args["limit"],
            offset=args["offset"],
            compute_facets=args["facets"],
            query=user_query,
            provider_code=provider_code,
            dataset_code=dataset_code,
            dimensions=dimensions,
            dimensions_codes_order=dimensions_codes_order,
        )

        if series_code is not None and series_response["response"]["numFound"] == 0:
            series_id = [provider_code, dataset_code, series_code]
            abort_json(404, args, "Series {!r} not found".format("/".join(series_id)))

        series_codes = sorted(pluck("code", series_response["response"]["docs"]))
        series_solr_by_id = {
            series_id_str_to_tuple(series_solr["id"]): series_solr
            for series_solr in series_response["response"]["docs"]
        }

    if args["format"] in TABULAR_FORMATS:
        return respond_download_spreadsheet_for_same_dataset(
            args, series_codes, dataset_dir, dataset_json, series_solr_by_id
        )

    if args["metadata"]:
        response_payload = {
            "provider": provider_solr,
            "dataset": build_dataset(dataset_json, dataset_solr),
        }
    else:
        response_payload = {}

    errors = []

    series_list, series_list_errors = build_series_list_for_dataset(
        dataset_dir, dataset_json, series_codes, series_solr_by_id, args["observations"]
    )
    if series_list_errors:
        errors.extend(series_list_errors)

    if args["align_periods"]:
        assert args["observations"]  # checked in query string parameters validation
        series_list = align_series_list(series_list)

    response_payload["series"] = build_paginated_data(
        [series_to_json(series) for series in series_list],
        args["limit"],
        series_response["response"]["numFound"],
        args["offset"],
    )

    series_response_facets = series_response.get("facets")
    if args["facets"] and series_response_facets:
        dimensions_values_labels = get_dimensions_values_labels(dataset_json)
        response_payload["series_dimensions_facets"] = build_dimensions_facets(
            series_response_facets, dimensions_codes_order, dimensions_values_labels
        )

    return make_json_response(
        {
            **response_payload,
            "errors": errors or None,
        },
        args,
    )
