# dbnomics-api -- DBnomics Web API
# By: DBnomics Team <christophe.benz@cepremap.org>
#
# Copyright (C) 2017, 2018 Cepremap
# https://git.nomics.world/dbnomics/dbnomics-api
#
# dbnomics-api is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# dbnomics-api is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Solr queries and utility functions."""

import logging
from urllib.parse import quote, unquote, urljoin

import requests
import ujson as json
from solrq import Q, QOperator, Value
from toolz import pipe, pluck
from toolz.curried import assoc, assoc_in, get_in, map, merge

from .utils import without_none_values

INTERNAL_FIELDS = ["_version_", "id", "type"]

log = logging.getLogger(__name__)


def compute_series_count_by_dataset_id_facet(query, limit, offset):
    """Compute facet with dataset ID because many datasets can have the same code and belong to different providers."""
    return solr_request(
        {
            "query": query,
            "filter": Q(type="series"),
            "facet": {
                "dataset_id": {
                    "type": "terms",
                    "field": "dataset_id",
                    "limit": limit,
                    "offset": offset,
                    "mincount": 1,
                },
                "dataset_count": "unique(dataset_id)",
            },
            "limit": 0,
        }
    )


def find_datasets(limit, offset, query=None, sort=None):
    return solr_request(
        without_none_values(
            {
                "query": query,
                "filter": Q(type="dataset"),
                "limit": limit,
                "offset": offset,
                "sort": sort,
            }
        )
    )


def find_datasets_by_provider_code(provider_code, dataset_code, limit, offset, sort):
    return solr_request(
        {
            "query": (
                Q(provider_code=provider_code, code=dataset_code)
                if dataset_code is not None
                else Q(provider_code=provider_code)
            ),
            "filter": Q(type="dataset"),
            "limit": limit,
            "offset": offset,
            "sort": sort,
        }
    )


def find_datasets_from_series(query, limit, offset):
    """Find series, build facets on `dataset_code`, then retrieve datasets."""
    facets_response = compute_series_count_by_dataset_id_facet(query, limit, offset)

    dataset_id_facet = facets_response["facets"].get("dataset_id")
    if not dataset_id_facet:
        return ([], 0)

    datasets_ids = list(pluck("val", dataset_id_facet["buckets"]))
    if not datasets_ids:
        return ([], 0)

    datasets_ids_q = Q(
        children=[Q(id=dataset_id) for dataset_id in datasets_ids], op=QOperator.or_
    )
    datasets_response = find_datasets(query=datasets_ids_q, limit=limit, offset=0)
    dataset_solr_by_id = {
        dataset_solr["id"]: dataset_solr
        for dataset_solr in datasets_response["response"]["docs"]
    }
    series_count_by_dataset_id = {
        bucket["val"]: bucket["count"] for bucket in dataset_id_facet["buckets"]
    }
    new_docs = pipe(
        [dataset_solr_by_id[dataset_id] for dataset_id in datasets_ids],
        map(
            lambda doc: assoc(
                doc, "nb_matching_series", series_count_by_dataset_id[doc["id"]]
            )
        ),
        list,
    )

    facets_count = facets_response["facets"]["dataset_count"]

    return (new_docs, facets_count)


def find_last_updated(type, limit, offset, sort):
    return solr_request(
        {
            "filter": Q(type=type),
            "limit": limit,
            "offset": offset,
            "sort": sort,
        }
    )


def find_providers(limit, offset, query=None):
    return solr_request(
        {
            "query": query,
            "filter": Q(type="provider"),
            "limit": limit,
            "offset": offset,
            "sort": "slug ASC",
        }
    )


def find_series(
    limit,
    offset=0,
    compute_facets=False,
    query=None,
    provider_code=None,
    dataset_code=None,
    dimensions=None,
    dimensions_codes_order=None,
    series_ids=None,
):
    facet = None
    if compute_facets:
        assert dimensions_codes_order is not None, dimensions_codes_order
        facet = {
            "dimension.{}".format(dimension_code): {
                "type": "terms",
                "field": "dimension_value_code.{}".format(quote(dimension_code)),
                "limit": -1,
                "mincount": 1,
                "domain": {"excludeTags": "DIMENSION.{}".format(quote(dimension_code))},
            }
            for dimension_code in dimensions_codes_order
        }

    query_children = []
    if query is not None:
        query_children.append(Q(_text_=query))
    if provider_code is not None:
        query_children.append(Q(provider_code=provider_code))
    if dataset_code is not None:
        query_children.append(Q(dataset_code=dataset_code))
    if query_children:
        query = Q(children=query_children)

    filter_ = [Q(type="series")]
    if dimensions is not None:
        filter_.extend(
            Q(
                children=[
                    Q(
                        **{
                            "{{!tag=DIMENSION.{0}}}dimension_value_code.{0}".format(
                                quote(dimension_code)
                            ): dimension_value_code
                        }
                    )
                    for dimension_value_code in dimensions_values_codes
                ],
                op=QOperator.or_,
            )
            for dimension_code, dimensions_values_codes in dimensions.items()
            if dimensions_values_codes
        )
    if series_ids:
        filter_.append(
            Q(
                children=[Q(id="/".join(series_id)) for series_id in series_ids],
                op=QOperator.or_,
            )
        )

    return solr_request(
        without_none_values(
            {
                "query": query,
                "filter": filter_,
                "facet": facet,
                "limit": limit,
                "offset": offset,
                "sort": "code ASC",
            }
        )
    )


def get_dataset(provider_code, dataset_code):
    response_json = solr_request(
        {
            "query": Q(provider_code=provider_code, code=dataset_code),
            "filter": Q(type="dataset"),
            "limit": 1,
        }
    )
    return get_in(["response", "docs", 0], response_json, default=None)


def get_provider(provider_code):
    response_json = solr_request(
        {
            "query": Q(code=provider_code),
            "filter": Q(type="provider"),
            "limit": 1,
        }
    )
    return get_in(["response", "docs", 0], response_json, default=None)


def iter_series_solr_dimensions(series_solr):
    prefix = "dimension_value_code."
    for k, dimension_value_code in series_solr.items():
        if k.startswith(prefix):
            dimension_code = unquote(k[len(prefix) :])
            yield (dimension_code, dimension_value_code)


def solr_request(json_payload):
    from .app import app  # Delay import to avoid cyclic imports.

    solr_query_url = urljoin(
        app.config["SOLR_BASE_URL"],
        "/solr/{}/query".format(app.config["SOLR_CORE_NAME"]),
    )

    # Deal with query value that can be simple text or solrq.Q instance
    query = json_payload.get("query")
    if query is None:
        query = Value("*", safe=True)
    if not isinstance(query, Q) and not isinstance(query, Value):
        raise ValueError("Unsupported query type: {!r}".format(query))

    filter_ = json_payload.get("filter")
    if filter_ is not None:
        if isinstance(filter_, list):
            filter_ = list(map(str, filter_))
        elif isinstance(filter_, Q):
            filter_ = str(filter_)
        else:
            raise ValueError("Unsupported filter type: {!r}".format(filter_))

    new_json_payload = pipe(
        json_payload,
        lambda d: merge(
            d,
            without_none_values(
                {
                    "query": str(query),
                    "filter": filter_,
                }
            ),
        ),
        lambda d: assoc_in(
            d, ["params", "q.op"], "AND"
        ),  # StandardQueryParser parameter
    )
    log.debug("Requesting Solr URL %r", solr_query_url)
    log.debug(json.dumps(new_json_payload, indent=2))
    response = requests.post(solr_query_url, json=new_json_payload)
    response.raise_for_status()
    response_json = response.json()
    if not response.ok:
        log.error(json.dumps(response_json, indent=2))
        error_trace = get_in(["error", "trace"], response_json, default=None)
        if error_trace is not None:
            log.error(error_trace)

    if app.debug:
        log.debug(json.dumps(response_json, indent=2))

    return response_json
