# dbnomics-api -- DBnomics Web API
# By: DBnomics Team <christophe.benz@cepremap.org>
#
# Copyright (C) 2017, 2018 Cepremap
# https://git.nomics.world/dbnomics/dbnomics-api
#
# dbnomics-api is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# dbnomics-api is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Functions manipulating series with their observations in tabular format."""

import io
import zipfile

import pandas as pd
from toolz import groupby, pluck

from .json_responses import (
    align_series_list,
    build_series_list,
    build_series_list_for_dataset,
    get_series_id,
)

TABULAR_FORMATS = {"csv", "xlsx"}


def build_dataframe(series_list, format):
    """Build dataframe. Assumes that `series_list` has been processed with align_periods"""

    def get_series_id_str(series):
        return "/".join(get_series_id(series))

    def format_column_name(series):
        series_name = series.get("series_name")
        series_id = get_series_id_str(series)
        if not series_name:
            return series_id
        return "{} ({})".format(series_name, series_id)

    columns = ["period"] + [format_column_name(series) for series in series_list]
    return pd.DataFrame(data=iter_dataframe_rows(series_list), columns=columns)


def build_file_content(sheets, format, errors):
    if format == "csv":
        if len(sheets) == 1 and not errors:
            _, df = sheets[0]
            return (format, df.to_csv(index=False))
        else:
            with io.BytesIO() as bio:
                with zipfile.ZipFile(bio, "w") as zip_file:
                    for sheet_name, df in sheets:
                        zip_file.writestr(
                            "{}.csv".format(sheet_name), df.to_csv(index=False)
                        )
                    if errors:
                        zip_file.writestr(
                            "errors.txt",
                            "".join(
                                format_error_prefix(error) + error["message"] + "\n"
                                for error in errors
                            ),
                        )
                return ("zip", bio.getvalue())
    elif format == "xlsx":
        with io.BytesIO() as bio:
            with pd.ExcelWriter(bio, engine="xlsxwriter") as writer:
                for sheet_name, df in sheets:
                    df.to_excel(writer, sheet_name=sheet_name, index=False)

                    # Generate a chart for series.
                    chart = writer.book.add_chart({"type": "line"})
                    max_nb_series = 10
                    for column_index, column_name in enumerate(
                        df.columns[1 : max_nb_series + 1], start=1
                    ):
                        chart.add_series(
                            {
                                # [sheetname, first_row, first_col, last_row, last_col]
                                "categories": [sheet_name, 1, 0, len(df), 0],
                                "values": [
                                    sheet_name,
                                    1,
                                    column_index,
                                    len(df),
                                    column_index,
                                ],
                                "name": column_name,
                            }
                        )

                    if len(df.columns) <= 11:  # column 1 is period
                        # One series: add a chart on the series sheet, sightly centered.
                        current_sheet = writer.sheets[sheet_name]
                        current_sheet.insert_chart(5, len(df.columns) + 3, chart)
                    else:
                        # Many series: add a "chart-sheet" with a chart containing many lines.
                        chartsheet = writer.book.add_chartsheet(
                            "{} chart".format(sheet_name)
                        )
                        if len(df.columns) > max_nb_series + 1:
                            chart.set_title(
                                {"name": "{} first series".format(max_nb_series)}
                            )
                        chartsheet.set_chart(chart)

                if errors:
                    errors_df = pd.DataFrame(
                        data=[
                            [format_error_prefix(error), error["message"]]
                            for error in errors
                        ]
                    )
                    errors_df.to_excel(
                        writer, sheet_name="errors", index=False, header=False
                    )

                writer.save()
            return (format, bio.getvalue())
    else:
        raise NotImplementedError(format)
    assert False, "Should never reach this line"


def iter_dataframe_rows(series_list):
    """Yield rows of multi-series tabular sheets, without the header.
    The first column is the period, and the others are observations.

    This function merges the periods and values of all the series, and assumes that series periods and values are already completed.

    >>> list(iter_dataframe_rows([{'period': ['2000'], 'value': [1]}]))
    [['2000', 1]]
    >>> list(iter_dataframe_rows([{'period': ['2000', '2001'], 'value': [1, 2]}]))
    [['2000', 1], ['2001', 2]]
    >>> list(iter_dataframe_rows([{'period': ['1999', '2000', '2001'], 'value': ['NA', 1, 2]}, {'period': ['1999', '2000', '2001'], 'value': [0, 'NA', 5]}]))
    [['1999', 'NA', 0], ['2000', 1, 'NA'], ['2001', 2, 5]]
    """
    # Because periods are already completed, they must be equal, and we can take the first one as reference.
    periods = series_list[0]["period"]
    values_list = list(pluck("value", series_list))
    for period, values in zip(periods, zip(*values_list)):
        yield [period] + list(values)


def iter_sheets(series_list, format):
    series_list_by_frequency = groupby(
        lambda series: series.get("@frequency"), series_list
    )
    for frequency, series_group in series_list_by_frequency.items():
        sheet_name = frequency.value if frequency is not None else "unknown frequency"
        yield (sheet_name, build_dataframe(series_group, format))


def respond_download_spreadsheet(sheets, format, base_file_name, errors):
    real_format, data = build_file_content(sheets, format, errors)
    mimetype_by_format = {
        "csv": "text/csv",
        "xlsx": "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
        "zip": "application/zip",
    }
    response_headers = {
        "Content-Disposition": 'attachment; filename="{}"'.format(
            "{}.{}".format(base_file_name, real_format)
        ),
        "Content-Type": mimetype_by_format[real_format],
    }
    return (data, response_headers)


def respond_download_spreadsheet_for_different_datasets(args, series_solr_by_id):
    errors = []
    _, _, series_list, series_list_errors = build_series_list(
        args["series_ids"], series_solr_by_id, observations=True
    )
    if series_list_errors:
        errors.extend(series_list_errors)
    series_list = align_series_list(series_list)
    sheets = list(iter_sheets(series_list, args["format"]))
    if not sheets:
        errors.append({"message": "Could not produce any series"})

    base_file_name = (
        "-".join(args["series_ids"][0])
        if len(args["series_ids"]) == 1
        else "DBnomics time series"
    )
    return respond_download_spreadsheet(sheets, args["format"], base_file_name, errors)


def respond_download_spreadsheet_for_same_dataset(
    args, series_codes, dataset_dir, dataset_json, series_solr_by_id
):
    errors = []
    series_list, series_list_errors = build_series_list_for_dataset(
        dataset_dir, dataset_json, series_codes, series_solr_by_id, observations=True
    )
    if series_list_errors:
        errors.extend(series_list_errors)
    series_list = align_series_list(series_list)
    sheets = list(iter_sheets(series_list, args["format"]))
    if not sheets:
        errors.append({"message": "Could not produce any series"})

    provider_code = dataset_dir.storage.provider_code
    dataset_code = dataset_dir.dataset_code
    base_file_name = (
        "-".join(dataset_dir.series_id(series_codes[0]))
        if len(series_codes) == 1
        else "{}-{} time series".format(provider_code, dataset_code)
    )  # Avoid "/" in file name.
    return respond_download_spreadsheet(sheets, args["format"], base_file_name, errors)


# ERRORS


def format_error_prefix(error):
    provider_code = error.get("provider_code")
    dataset_code = error.get("dataset_code")
    series_code = error.get("series_code")
    if (
        provider_code is not None
        and dataset_code is not None
        and series_code is not None
    ):
        return "Series {}/{}/{}: ".format(provider_code, dataset_code, series_code)
    if provider_code is not None and dataset_code is not None:
        return "Dataset {}/{}: ".format(provider_code, dataset_code)
    if provider_code is not None:
        return "Provider {}: ".format(provider_code)
    return ""
