# dbnomics-api -- DBnomics Web API
# By: DBnomics Team <christophe.benz@cepremap.org>
#
# Copyright (C) 2017, 2018 Cepremap
# https://git.nomics.world/dbnomics/dbnomics-api
#
# dbnomics-api is free software; you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# dbnomics-api is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


"""Functions related to input validation."""

import ujson as json
import valideer as V

from . import tabular_responses
from .json_errors import abort_json


def csv_list(v):
    """Return a `list` of `str` from a `str` containing comma-separated values.

    >>> csv_list("")
    []
    >>> csv_list("a")
    ['a']
    >>> csv_list(" a ")
    [' a ']
    >>> csv_list("a,")
    ['a']
    >>> csv_list("a,,")
    ['a']
    >>> csv_list(",a")
    ['', 'a']
    >>> csv_list("a,b")
    ['a', 'b']
    >>> csv_list("a,b,")
    ['a', 'b']
    """
    if v == "":
        return []
    separator = ","
    return v.rstrip(separator).split(separator)


def guess_bool(value: str):
    if value.lower() in {"true", "yes", "on", "1"}:
        return True
    return False


check_bool = V.Nullable(V.AdaptBy(guess_bool), default=False)

check_is_positive = V.Range(V.AdaptTo(int), min_value=0)

check_dimensions = V.Nullable(
    V.ChainOf(
        V.AdaptBy(lambda s: json.loads(s) if s else {}),
        V.Mapping(key_schema=str, value_schema=[str]),
    ),
    default={},
)

check_offset = V.Nullable(check_is_positive, default=0)


def check_series_ids(max_length):
    return V.Nullable(
        V.ChainOf(
            V.AdaptBy(csv_list),
            V.HomogeneousSequence(
                V.AdaptBy(series_id_str_to_tuple), min_length=1, max_length=max_length
            ),
        ),
        default=[],
    )


def series_id_str_to_tuple(series_id_str):
    """Example:
    >>> series_id_str_to_tuple('a/b/c')
    ('a', 'b', 'c')
    """
    error_message = "Series ID {!r} must be formatted like 'provider_code/dataset_code/series_code'".format(
        series_id_str
    )
    if series_id_str.count("/") != 2:
        raise V.ValidationError(error_message)
    provider_code, dataset_code, series_code = [
        s.strip() for s in series_id_str.split("/", 2)
    ]
    if not provider_code or not dataset_code or not series_code:
        raise V.ValidationError(error_message)
    return (provider_code, dataset_code, series_code)


def convert_args_or_abort(args, schema):
    with V.parsing(additional_properties=False):
        return validate_or_abort(args, V.parse(schema), args.to_dict())


def validate_or_abort(args, validator, value):
    try:
        return validator.validate(value)
    except V.ValidationError as exc:
        abort_json(400, args, message=str(exc))


series_observations_args = {
    "observations": check_bool,
    "align_periods": check_bool,
    "format": V.Nullable(V.Enum(["csv", "json", "xlsx"]), default="json"),
}


def check_series_observations_args(args):
    if args["format"] in tabular_responses.TABULAR_FORMATS:
        args = {**args, "observations": True}
    if not args["observations"]:
        for arg_name in ["align_periods"]:
            arg = args.get(arg_name)
            if arg:
                abort_json(
                    400,
                    args,
                    "'{}' parameter must be used with 'observations' parameter".format(
                        arg_name
                    ),
                )
    return args
